const ds = require('./src/dataspec')

const isString = ds.spec(x => typeof x === 'string')
  .examples('foo', 'bar', 'baz', 'qux', 'quux', 'hello world')

const isNull = ds.spec(x => x === null).examples(null)

const isBoolean = ds.spec(x => x === true || x === false)
  .examples(true, false)

const isNumber = ds.spec(x => typeof x === 'number')
  .examples(0.3, -8.912, 15, -7.4, -18, -5, 0, 0.3, 5.4,
    42, 46, 55.82, 259, 10237, 123790, -1298.23)

const isInteger = isNumber.and(Number.isInteger)

const isPositive = isNumber.and(x => x > 0)

const isZero = ds.spec(x => x === 0).examples(0)

const isDateString = isString.and(x => /\d{4}-\d{2}-\d{2}/.test(x))
  .examples('2017-02-15', '2015-12-01', '2017-01-30', '2017-03-06')

const isNumericString = isString.and(x => /\d+(\.\d+)?/.test(x))
  .examples('20', '8.5', '-10.7', '0.4', '-20.7')

const isIntegerString = isString.and(x => /\d+/.test(x))
  .examples('20', '8', '107', '437298')

const isBase64 = isString.and(x => /^data:image\/png;base64,/.test(x))
  .examples("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAABkAAAAEsCAYAAACbnX8IAAAg",
    "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAABkAAAAEsCAYAAACbnX8IAAAgAElEQVR4Xu")

const isEmail = isString.and(x => /.+@.+\..+/.test(x))
  .examples('foo@example.com', 'bar@example.com', 'baz@example.com', 'qux@example.com')

const isPetBag = ds.objectOf({
  "sealNumber": isString,
  "customerNumber": isString,
  "customerId": isString,
  "customerName": isString,
  "customerHomeStore": isString
})

const isTransportGood = ds.objectOf({
  "id": isString,
  "balance": isInteger,
  "delivered": isInteger.and(isPositive).or(isZero),
  "description": isString,
  "value": isNumber.and(isPositive).or(isZero),
  "returned": isInteger.and(isPositive).or(isZero)
})

const isOrder = ds.objectOf({
  "deliveryComment": isString,
  "orderId": isString,
  "orderNumber": isString,
  "orderDate": isString,
  "paymentType": isString,
  "grossTotal": isNumber.and(isPositive).or(isZero)
})

const isUnitReturn = ds.objectOf({
  "description": isString,
  "deliveryDate": isDateString,
  "reason": isString,
  "subsysNumber": isString,
  "mgbNumber": isString,
  "quantity": isInteger.and(isPositive)
})

const isWeightedReturn = ds.objectOf({
  "description": isString,
  "deliveryDate": isDateString,
  "reason": isString,
  "subsysNumber": isString,
  "mgbNumber": isString,
  "quantity": isNumber.and(isPositive)
})

const isCustomerEmpty = ds.objectOf({
  "description": isString,
  "quantity": isInteger.and(isPositive).or(isZero),
  "unitValue": isNumber.and(isPositive).or(isZero),
  "ean": isIntegerString
})

const deliverySummary = ds.objectOf({
  "street": isString,
  "houseNumber": isString.or(isNull),
  "buildingName": isString.or(isNull),
  "zipCode": isString.or(isNull),
  "city": isString,
  "customerNumber": isString,
  "customerName": isString,
  "unloadingTemperatureFresh": isNumericString,
  "unloadingTemperatureFrozen": isNumericString,
  "customerSignature": isBase64,
  "driverSignature": isBase64,
  "driverSignatoryName": isString,
  "cashToCollect": isNumber.and(isPositive).or(isZero),
  "creditTotal": isNumber.and(isPositive).or(isZero),
  "collectedCash": isNumber.and(isPositive).or(isZero),
  "petEmptiesBags": ds.arrayOf(isPetBag),
  "transportGoods": ds.arrayOf(isTransportGood),
  "orders": ds.arrayOf(isOrder),
  "unitReturns": ds.arrayOf(isUnitReturn),
  "weightedReturns": ds.arrayOf(isWeightedReturn),
  "customerEmpties": ds.arrayOf(isCustomerEmpty),
  "loadingTemperatureFresh": isNumber,
  "loadingTemperatureFrozen": isNumber,
  "storeEmail": isEmail,
  "storePhoneNumber": isString,
  "storeFaxNumber": isString.or(isNull),
  "storeUrl": isString,
  "storePrintName": isString,
  "storeFooterLeft": isString,
  "storeFooterMiddle": isString,
  "storeFooterRight": isString,
  "storeCurrencyCode": isString,
  "countryConfiguration": ds.objectOf({
    "recordTemperatures": isBoolean,
    "displayEmptiesTotal": isBoolean,
    "displayTransportGoodsTotal": isBoolean,
    "displayNetTotal": isBoolean,
    "displayReturnsTotal": isBoolean,
    "allowCardPayment": isBoolean,
    "displayTransportGoodsOverview": isBoolean
  })
})

module.exports = deliverySummary
